import Base from './base';
import { MARGIN_DISTANCE } from '../../constants'
import { getActualPoint, getClickPoint, createSvg, getSvgPathFromStroke, getInitialPoint } from './utils';
import { onClickOutside, } from '../ui_utils'
import { getStroke } from "perfect-freehand";

export default class Ink extends Base {
  
  constructor ({
    container,
    annotation,
    page,
    viewport,
    scale,
    eventBus,
    layer,
    show = false
  }) {
    super({
      container,
      annotation,
      page,
      viewport,
      scale,
      eventBus,
      show
    })

    this.layer = layer
    this.hidden = true
    this.initial = false
    this.outline = null

    this.start = null
    this.end = null
    this.rect = null

    this.newStart = null
    this.newEnd = null

    this.svgElement = null

    this.startCircle = null
    this.endCircle = null
    this.startState = null

    this.show = show

    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)
    this.onDelete = this.handleDelete.bind(this)
    this.onKeydown = this.handleKeydown.bind(this)
    this.path = []
    this.render()
  }

  getFreeDrawSvgPath () {
    const options = {
      simulatePressure: false,
      size: 1,
      thinning: 0,
      smoothing: 0.3,
      streamline: 0.5,
      easing: (t) => Math.sin((t * Math.PI) / 2),
      last: false
    }
    return getSvgPathFromStroke(getStroke(this.path, options))
  }
  
  getRect () {
    const xArray = [];
    const yArray = [];
    this.path.forEach((point) => {
      xArray.push(point[0]);
      yArray.push(point[1]);
    });
    const top = Math.min(...yArray);
    const left = Math.min(...xArray);
    const bottom = Math.max(...yArray);
    const right = Math.max(...xArray);
    this.path = this.path.map((ele) => {
      let x = ele[0] - left
      let y = ele[1] - top
      return [x, y]
    })
    return {
      left,
      top,
      bottom,
      right,
      width: right - left,
      height: bottom - top
    }
  }

  formatePoint (x, y) {
    const point = getActualPoint(
      {
        x,
        y
      },
      this.viewport,
      this.scale
    )
    return [point.x, point.y]
  }

  render () {
    let gesture
    if (!this.annotation.inklist.gesture) {
      gesture = this.annotation.inklist
    } else {
      gesture = this.annotation.inklist.gesture
    }

    gesture = gesture.split(';')

    const length = gesture.length
    for (let i = 0; i < length; i++) {
      const point = gesture[i].split(',')
      this.path.push(this.formatePoint(point[0], point[1]))
    }

    const rect = this.getRect()
    this.rect = rect

    const annotation = this.annotation

    this.start = {
      x: rect.left,
      y: rect.bottom
    }

    this.end = {
      x: rect.right,
      y: rect.top
    }

    const actualbdwidth = (this.annotation.lineWidth || this.annotation.width) * this.scale
    
    const annotationContainer = document.createElement('div')
    annotationContainer.id = annotation.name
    annotationContainer.className = 'annotation'
    annotationContainer.style.top = rect.top + 'px'
    annotationContainer.style.left = rect.left + 'px'
    annotationContainer.style.width = rect.width + 'px'
    annotationContainer.style.height = rect.height + 'px'
    this.annotationContainer = annotationContainer

    this.svgElement = document.createElementNS("http://www.w3.org/2000/svg", "svg");

    this.svgElement.setAttribute("viewBox", `0 0 ${rect.width} ${rect.height}`)
    this.svgElement.addEventListener('mousedown', this.handleClick.bind(this))

    let shapeElement = createSvg(
      "path",
      {
        d: this.getFreeDrawSvgPath(),
        stroke: annotation.borderColor || annotation.color,
        'stroke-width': actualbdwidth,
        'stroke-opacity': annotation.transparency,
        fill: 'none',
        'fill-opacity': 0,
      }
    );

    this.shapeElement = shapeElement

    this.svgElement.append(this.shapeElement)

    this.annotationContainer.append(this.svgElement)

    this.container.append(this.annotationContainer)

    this.outerLineContainer = document.createElement('div')
    this.outerLineContainer.className = 'outline-container'
    this.deletetButton = createSvg(
      "svg",
      {
        width: "30",
        height: "30",
        viewBox: "0 0 30 30",
        fill: "none",
        class: 'delete-button'
      },
      {
        width: '30px',
        height: '30px'
      }
    );
    this.deletetButton.addEventListener('click', this.onDelete)
    this.deletetButton.innerHTML = this.deleteSvgStr

    const left = (rect.left + rect.width - 30) + 'px'
    const top = (rect.top + rect.height + 8) + 'px'
    this.deletetButton.style.left = left
    this.deletetButton.style.top = top
    
    const outerLine = document.createElementNS("http://www.w3.org/2000/svg", "svg");

    outerLine.style.position = 'absolute'
    outerLine.style.zIndex = 1
    outerLine.style.left = `${rect.left - 7}px`
    outerLine.style.top = `${rect.top - 7}px`
    outerLine.style.width = `${rect.width + 7 * 2}px`
    outerLine.style.height = `${rect.height + 7 * 2}px`
    this.outerLine = outerLine
    
    this.moveRect = createSvg(
      "rect",
      {
        class: "move",
        'data-id': "move",
        stroke: "#1460F3",
        'stroke-width': 1,
        'fill-opacity': 0,
        width: rect.width + 8,
        height: rect.height + 8,
        x: 3,
        y: 3
      }
    );
    
    this.topLeftRect = createSvg(
      "rect",
      {
        class: "nw-resize",
        'data-id': "nw-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: 0,
        y: 0
      }
    );

    this.LeftRect = createSvg(
      "rect",
      {
        class: "w-resize",
        'data-id': "w-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: 0,
        y: rect.height / 2 + 4
      }
    );

    this.bottomLeftRect = createSvg(
      "rect",
      {
        class: "sw-resize",
        'data-id': "sw-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: 0,
        y: rect.height + 8
      }
    );

    this.bottomRect = createSvg(
      "rect",
      {
        class: "s-resize",
        'data-id': "s-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: rect.width / 2 + 4,
        y: rect.height + 8
      }
    );
    
    this.bottomRightRect = createSvg(
      "rect",
      {
        class: "se-resize",
        'data-id': "se-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: rect.width + 8,
        y: rect.height + 8
      }
    );

    this.rightRect = createSvg(
      "rect",
      {
        class: "e-resize",
        'data-id': "e-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: rect.width + 8,
        y: rect.height / 2 + 4
      }
    );

    this.topRightRect = createSvg(
      "rect",
      {
        class: "ne-resize",
        'data-id': "ne-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: rect.width + 8,
        y: 0
      }
    );

    this.topRect = createSvg(
      "rect",
      {
        class: "n-resize",
        'data-id': "n-resize",
        fill: "#1460F3",
        stroke: "none",
        width: 6,
        height: 6,
        x: rect.width / 2 + 4,
        y: 0
      }
    );

    this.outerLine.append(this.moveRect)
    this.outerLine.append(this.topLeftRect)
    this.outerLine.append(this.LeftRect)
    this.outerLine.append(this.bottomLeftRect)
    this.outerLine.append(this.bottomRect)
    this.outerLine.append(this.bottomRightRect)
    this.outerLine.append(this.rightRect)
    this.outerLine.append(this.topRightRect)
    this.outerLine.append(this.topRect)
    this.outerLineContainer.append(this.outerLine)
    this.outerLineContainer.append(this.deletetButton)
    this.initial = true
  }

  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    };
  };

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end });
    const actualbdwidth = (this.annotation.width || this.annotation.lineWidth) * this.scale

    return {
      top: initRect.top - actualbdwidth,
      left: initRect.left - actualbdwidth,
      width: initRect.width + actualbdwidth * 2,
      height: initRect.height + actualbdwidth * 2,
    }
  }

  updateTool () {
    if (!this.initial) return
    if (this.hidden) {
      if (this.layer.selectedElementName === this.annotation.name) {
        this.layer.selectedElementName = null
      }
      this.outerLineContainer.remove()
      document.removeEventListener('keydown', this.onKeydown)
    } else {
      if (this.layer.selectedElementName !== this.annotation.name) {
        this.layer.selectedElementName = this.annotation.name
      }
      this.container.append(this.outerLineContainer)
      this.outerLine.addEventListener('mousedown', this.onMousedown)
      this.outerLine.addEventListener('touchstart', this.onMousedown)
      document.addEventListener('keydown', this.onKeydown)
    }
  }

  handleClick (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    if (!this.hidden) return
    this.hidden = false
    this.updateTool()
    onClickOutside([this.svgElement, this.outerLine, this.deletetButton], this.handleOutside.bind(this))
  }

  handleOutside () {
    this.hidden = !this.hidden
    if (this.layer.selectedElementName === this.annotation.name) {
      this.layer.selectedElementName = null
    }
    this.outerLine.removeEventListener('mousedown', this.onMousedown)
    this.outerLine.removeEventListener('touchstart', this.onMousedown)
    this.outerLineContainer.remove()
  }


  handleMouseDown (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    const operatorId = event.target.getAttribute('data-id')
  
    const { pageX, pageY } = getClickPoint(event)
    this.startState = {
      operator: operatorId,
      clickX: pageX,
      clickY: pageY,
    }
    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)
    
  };

  handleMouseMove (event) {
    if (event.type === 'touchmove') {
      document.body.style.overscrollBehavior = 'none';
    }
    this.moving = true
    const { pageX, pageY } = getClickPoint(event)
    const start = this.start
    const end = this.end
    const startState = this.startState

    const { width, height } = this.viewport
    if (startState.operator === 'nw-resize') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.end.y);
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(start.y - 5, y1)
      y1 = Math.max(0, y1)

      this.update({
        start: { x: x1, y: start.y },
        end: { x: end.x, y: y1 },
      });
    } else if (startState.operator === 'w-resize') {
      let x1 = pageX - (startState.clickX - this.start.x);
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)

      this.update({
        start: { x: x1, y: start.y },
        end,
      });
    } else if (startState.operator === 'sw-resize') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.start.y);
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(height, y1)
      y1 = Math.max(end.y + 5, y1)

      this.update({
        start: { x: x1, y: y1 },
        end,
      });
    } else if (startState.operator === 's-resize') {
      let y1 = pageY - (startState.clickY - this.start.y);
      y1 = Math.min(height, y1)
      y1 = Math.max(end.y + 5, y1)
      this.update({
        start: { x: start.x, y: y1 },
        end,
      });
    } else if (startState.operator === 'se-resize') {
      let x = pageX - (startState.clickX - this.end.x);
      let y = pageY - (startState.clickY - this.start.y);
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      y = Math.min(height, y)
      y = Math.max(end.y + 5, y)
      this.update({
        start: { x: start.x, y },
        end: { x, y: end.y },
      });
    } else if (startState.operator === 'e-resize') {
      let x = pageX - (startState.clickX - this.end.x);
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      this.update({
        start,
        end: { x, y: end.y },
      });
    } else if (startState.operator === 'ne-resize') {
      let x = pageX - (startState.clickX - this.end.x);
      let y = pageY - (startState.clickY - this.end.y);
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      y = Math.min(start.y - 5, y)
      y = Math.max(0, y)
      this.update({
        start,
        end: { x, y },
      });
    } else if (startState.operator === 'n-resize') {
      let y = pageY - (startState.clickY - this.end.y);
      y = Math.min(start.y - 5, y)
      y = Math.max(0, y)
      this.update({
        start,
        end: { x: end.x, y },
      });
    } else if (startState.operator === 'move') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.start.y);
      let x2 = pageX - (startState.clickX - this.end.x);
      let y2 = pageY - (startState.clickY - this.end.y);

      const rect = {
        width: Math.abs(start.x -end.x),
        height: Math.abs(start.y -end.y)
      }
      if (x1 < x2) {
        x1 = Math.max(0, x1)
        x1 = Math.min(width - rect.width, x1)
        x2 = Math.max(rect.width, x2)
        x2 = Math.min(width, x2)
      } else {
        x2 = Math.max(0, x2)
        x2 = Math.min(width - rect.width, x2)
        x1 = Math.max(rect.width, x1)
        x1 = Math.min(width, x1)
      }

      if (y1 < y2) {
        y1 = Math.max(0, y1)
        y1 = Math.min(height - rect.height, y1)
        y2 = Math.max(rect.height, y2)
        y2 = Math.min(height, y2)
      } else {
        y2 = Math.max(0, y2)
        y2 = Math.min(height - rect.height, y2)
        y1 = Math.max(rect.height, y1)
        y1 = Math.min(height, y1)
      }

      this.update({
        start: { x: x1, y: y1 },
        end: { x: x2, y: y2 },
      });
    }
  }

  handleDelete (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    this.handleOutside()
    this.annotationContainer.remove()
    this.annotation.isDelete = true
    this.eventBus.dispatch('annotationChange', {
      type: 'delete',
      annotation: {
        operate: "del-annot",
        name: this.annotation.name,
        page: this.page
      }
    })
  }

  handleKeydown (event) {
    if (event.key === 'Delete' || event.key === 'Backspace') {
      this.handleDelete()
    }
  }

  handleMouseUp (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    if (event.type === 'touchend') {
      document.body.style.overscrollBehavior = 'auto';
    }
    this.moving = false
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
    this.start = this.newStart
    this.end = this.newEnd

    this.startState = null

    const inklist = this.formatPath()
    const initialStart = getInitialPoint(this.start, this.viewport, this.scale)
    const initialEnd = getInitialPoint(this.end, this.viewport, this.scale)
    const rect = initialStart.x + ',' + initialStart.y + ',' + initialEnd.x + ',' + initialEnd.y

    this.eventBus.dispatch('annotationChange', {
      type: 'modify',
      annotation: {
        operate: "mod-annot",
        name: this.annotation.name,
        page: this.page,
        inklist,
        rect
      }
    })
  }

  handlePoint (point) {
    const rect = this.rect
    const newPoint = getInitialPoint(
      {
        x: point[0] + rect.left,
        y: point[1] + rect.top
      },
      this.viewport,
      this.scale
    )
    return newPoint.x + ',' + newPoint.y
  }

  formatPath () {
    let path = ''
    const length = this.path.length
    for (let i = 0; i < length; i = i+2) {
      path+=(this.handlePoint(this.path[i]))
      if (i + 1 < length) {
        path += (';' + this.handlePoint(this.path[i + 1]))
        if (i + 2 < length) {
          path += ';'
        }
      }
    }
    return path
  }

  update ({ start, end }) {
    const rect = this.calculate(start, end)
    if (this.startState && this.startState.operator !== 'move') {
      this.path = this.path.map((ele) => {
        const x = (rect.width / this.rect.width) * ele[0]
        const y = (rect.height / this.rect.height) * ele[1]
        return [x, y]
      })
    }

    this.rect = rect
    this.newStart = start
    this.newEnd = end

    const actualbdwidth = (this.annotation.lineWidth || this.annotation.width) * this.scale

    this.annotationContainer.style.top = rect.top + 'px'
    this.annotationContainer.style.left = rect.left + 'px'
    this.annotationContainer.style.width = rect.width + 'px'
    this.annotationContainer.style.height = rect.height + 'px'

    this.svgElement.setAttribute("viewBox", `0 0 ${rect.width} ${rect.height}`)

    this.shapeElement.setAttribute("width", rect.width - actualbdwidth);
    this.shapeElement.setAttribute("height", rect.height - actualbdwidth);
    this.shapeElement.setAttribute("d", this.getFreeDrawSvgPath());

    this.outerLine.style.left = `${rect.left - 7}px`
    this.outerLine.style.top = `${rect.top - 7}px`
    this.outerLine.style.width = `${rect.width + 7 * 2}px`
    this.outerLine.style.height = `${rect.height + 7 * 2}px`

    const left = (rect.left + rect.width - 30) + 'px'
    const top = (rect.top + rect.height + 8) + 'px'
    this.deletetButton.style.left = left
    this.deletetButton.style.top = top
    
    this.moveRect.setAttribute('width',  rect.width + 8)
    this.moveRect.setAttribute('height',  rect.height + 8)
    
    this.LeftRect.setAttribute("y", rect.height / 2 + 4)

    this.bottomLeftRect.setAttribute("y", rect.height + 8)

    this.bottomRect.setAttribute("x", rect.width / 2 + 4)
    this.bottomRect.setAttribute("y", rect.height + 8)



    this.bottomRightRect.setAttribute("x", rect.width + 8)
    this.bottomRightRect.setAttribute("y", rect.height + 8)


    this.rightRect.setAttribute("x", rect.width + 8,)
    this.rightRect.setAttribute("y", rect.height / 2 + 4)

    
    this.topRightRect.setAttribute("x", rect.width + 8)

    this.topRect.setAttribute("x", rect.width / 2 + 4)
  }
}
