import Line from './line.js'
import Shape from './shape.js'
import Ink from './ink.js'
import Freetext from './freetext.js'
import Text from './text.js'
import PaintFreetext from './paint/freetext'
import PaintShape from './paint/shape';
import PaintText from './paint/text';
import TextField from '../form/text_field.js'
import AddTextField from '../form/add_text_field.js'
import CheckBox from '../form/check_box.js'
import AddCheckBox from '../form/add_check_box.js'
import RadioButton from '../form/radio_button.js'
import AddRadioButton from '../form/add_radio_button.js'
import PushButton from '../form/push_button.js'
import AddPushButton from '../form/add_push_button.js'

class ComPDFAnnotationLayer {
  constructor({
    annotationStore,
    pageViewer,
    pageDiv,
    pdfPage,
    annotations = null,
    annotationsAll,
    viewport,
    scale,
    pageIndex,
    eventBus
  }) {
    this._cancelled = false;
    this.pageViewer = pageViewer
    this.pageDiv = pageDiv;
    this.pdfPage = pdfPage;
    this.annotations = annotations || [];
    this.annotationsAll = annotationsAll;
    this.viewport = viewport
    this.annotationsArray = []
    this.scale = scale
    this.eventBus = eventBus
    this.page = pageIndex
    this.tool = pageViewer.tool || ''
    this.color = pageViewer.color || ''
    this.annotateManager = null
    this.isDraging = false
    this.selectedElementName = null
    this.addMode = false
    this.toolMode = pageViewer.toolMode || ''

    this.annotationStore = annotationStore

    this.onHandleTool = this.handleTool.bind(this)
    this.onHandleToolMode = this.handleToolMode.bind(this)
    this.onHandlePropertyPanelChanged = this.handlePropertyPanelChanged.bind(this)
  }

  handleTool({
    tool,
    color
  }) {
    this.tool = tool
    this.color = color
    this.pageViewer.tool = tool
    this.pageViewer.color = color
    if (!tool) {
      document.querySelector('.document').classList.remove('annotation-edit')
    }
    const markups = ['highlight', 'underline', 'squiggly', 'strikethrough']
    const shapes = ['line', 'arrow', 'circle', 'square', 'ink']
    if (markups.includes(tool) && color) {
      this.svgElement.classList.add('point-none')
    }
    if (this.annotateManager) {
      this.annotateManager.tool = tool
      this.annotateManager.color = color
    }
    if (this.freetextManager) {
      this.freetextManager.tool = tool
      this.freetextManager.color = color
    }
    if (this.textManager) {
      this.textManager.tool = tool
      this.textManager.color = color
    }
    if (this.textFieldManager) {
      this.textFieldManager.tool = tool
      this.textFieldManager.color = color
    }
    if (this.checkBoxManager) {
      this.checkBoxManager.tool = tool
      this.checkBoxManager.color = color
    }
    if (this.radioButtonManager) {
      this.radioButtonManager.tool = tool
      this.radioButtonManager.color = color
    }
    if (this.pushButtonManager) {
      this.pushButtonManager.tool = tool
      this.pushButtonManager.color = color
    }
    if (shapes.includes(tool) && color) {
      document.querySelector('.document').classList.add('annotation-edit')
      if (this.freetextManager) {
        this.freetextManager.reset()
      }
      if (this.textManager) {
        this.textManager.reset()
      }
      if (this.textFieldManager) {
        this.textFieldManager.reset()
      }
      if (this.checkBoxManager) {
        this.checkBoxManager.reset()
      }
      if (this.radioButtonManager) {
        this.radioButtonManager.reset()
      }
      if (this.pushButtonManager) {
        this.pushButtonManager.reset()
      }
      if (this.annotateManager) {
        this.annotateManager.init()
        return
      }
      this.annotateManager = new PaintShape({
        tool,
        color,
        container: this.pageDiv,
        svgElement: this.svgElement,
        layer: this,
        viewport: this.viewport,
        scale: this.scale,
        page: this.page,
        eventBus: this.eventBus,
        selectedElementName: this.selectedElementName
      })
    } else if (tool === 'freetext') {
      document.querySelector('.document').classList.add('annotation-edit')
      if (this.annotateManager) {
        this.annotateManager.reset()
      }
      if (this.textManager) {
        this.textManager.reset()
      }
      if (this.textFieldManager) {
        this.textFieldManager.reset()
      }
      if (this.checkBoxManager) {
        this.checkBoxManager.reset()
      }
      if (this.radioButtonManager) {
        this.radioButtonManager.reset()
      }
      if (this.pushButtonManager) {
        this.pushButtonManager.reset()
      }
      if (this.freetextManager) {
        this.freetextManager.init()
        return
      }
      this.freetextManager = new PaintFreetext({
        tool,
        color,
        container: this.pageDiv,
        svgElement: this.svgElement,
        layer: this,
        viewport: this.viewport,
        scale: this.scale,
        page: this.page,
        eventBus: this.eventBus,
        selectedElementName: this.selectedElementName
      })
    } else if (tool === 'text') {
      document.querySelector('.document').classList.add('annotation-edit')
      if (this.annotateManager) {
        this.annotateManager.reset()
      }
      if (this.freetextManager) {
        this.freetextManager.reset()
      }
      if (this.textFieldManager) {
        this.textFieldManager.reset()
      }
      if (this.checkBoxManager) {
        this.checkBoxManager.reset()
      }
      if (this.radioButtonManager) {
        this.radioButtonManager.reset()
      }
      if (this.pushButtonManager) {
        this.pushButtonManager.reset()
      }
      if (this.textManager) {
        this.textManager.init()
        return
      }
      this.textManager = new PaintText({
        tool,
        color,
        container: this.pageDiv,
        svgElement: this.svgElement,
        layer: this,
        viewport: this.viewport,
        scale: this.scale,
        page: this.page,
        eventBus: this.eventBus,
        selectedElementName: this.selectedElementName
      })
    } else if (tool === 'textfield') {
      document.querySelector('.document').classList.add('annotation-edit')
      if (this.annotateManager) {
        this.annotateManager.reset()
      }
      if (this.freetextManager) {
        this.freetextManager.reset()
      }
      if (this.textManager) {
        this.textManager.reset()
      }
      if (this.checkBoxManager) {
        this.checkBoxManager.reset()
      }
      if (this.radioButtonManager) {
        this.radioButtonManager.reset()
      }
      if (this.pushButtonManager) {
        this.pushButtonManager.reset()
      }
      if (this.textFieldManager) {
        this.textFieldManager.init()
        return
      }
      this.textFieldManager = new AddTextField({
        tool,
        color,
        container: this.pageDiv,
        svgElement: this.svgElement,
        layer: this,
        viewport: this.viewport,
        scale: this.scale,
        page: this.page,
        eventBus: this.eventBus,
        selectedElementName: this.selectedElementName
      })
    } else if (tool === 'checkbox') {
      document.querySelector('.document').classList.add('annotation-edit')
      if (this.annotateManager) {
        this.annotateManager.reset()
      }
      if (this.freetextManager) {
        this.freetextManager.reset()
      }
      if (this.textManager) {
        this.textManager.reset()
      }
      if (this.textFieldManager) {
        this.textFieldManager.reset()
      }
      if (this.radioButtonManager) {
        this.radioButtonManager.reset()
      }
      if (this.pushButtonManager) {
        this.pushButtonManager.reset()
      }
      if (this.checkBoxManager) {
        this.checkBoxManager.init()
        return
      }
      this.checkBoxManager = new AddCheckBox({
        tool,
        color,
        container: this.pageDiv,
        svgElement: this.svgElement,
        layer: this,
        viewport: this.viewport,
        scale: this.scale,
        page: this.page,
        eventBus: this.eventBus,
        selectedElementName: this.selectedElementName
      })
    } else if (tool === 'radiobutton') {
      document.querySelector('.document').classList.add('annotation-edit')
      if (this.annotateManager) {
        this.annotateManager.reset()
      }
      if (this.freetextManager) {
        this.freetextManager.reset()
      }
      if (this.textManager) {
        this.textManager.reset()
      }
      if (this.textFieldManager) {
        this.textFieldManager.reset()
      }
      if (this.checkBoxManager) {
        this.checkBoxManager.reset()
      }
      if (this.pushButtonManager) {
        this.pushButtonManager.reset()
      }
      if (this.radioButtonManager) {
        this.radioButtonManager.init()
        return
      }
      this.radioButtonManager = new AddRadioButton({
        tool,
        color,
        container: this.pageDiv,
        svgElement: this.svgElement,
        layer: this,
        viewport: this.viewport,
        scale: this.scale,
        page: this.page,
        eventBus: this.eventBus,
        selectedElementName: this.selectedElementName
      })
    } else if (tool === 'pushbutton') {
      document.querySelector('.document').classList.add('annotation-edit')
      if (this.annotateManager) {
        this.annotateManager.reset()
      }
      if (this.freetextManager) {
        this.freetextManager.reset()
      }
      if (this.textManager) {
        this.textManager.reset()
      }
      if (this.textFieldManager) {
        this.textFieldManager.reset()
      }
      if (this.checkBoxManager) {
        this.checkBoxManager.reset()
      }
      if (this.radioButtonManager) {
        this.radioButtonManager.reset()
      }
      if (this.pushButtonManager) {
        this.pushButtonManager.init()
        return
      }
      this.pushButtonManager = new AddPushButton({
        tool,
        color,
        container: this.pageDiv,
        svgElement: this.svgElement,
        layer: this,
        viewport: this.viewport,
        scale: this.scale,
        page: this.page,
        eventBus: this.eventBus,
        selectedElementName: this.selectedElementName
      })
    }
  }
  render(viewport) {
    if (this._cancelled) {
      return;
    }
    this.eventBus._on('toolChanged', this.onHandleTool)
    this.eventBus._on('toolModeChanged', this.onHandleToolMode)
    this.eventBus._on('setDefaultSelect', this.onHandleDefaultSelect)
    this.eventBus._on('propertyPanelChanged', this.onHandlePropertyPanelChanged)

    if (viewport) {
      const clonedViewport = viewport.clone({ dontFlip: true });
      this.viewport = clonedViewport;
      this.scale = clonedViewport.scale

      if (this.annotateManager) {
        this.annotateManager.viewport = clonedViewport
      }
      if (this.freetextManager) {
        this.freetextManager.viewport = clonedViewport
      }
      if (this.textManager) {
        this.textManager.viewport = clonedViewport
      }
      if (this.textFieldManager) {
        this.textFieldManager.viewport = clonedViewport
      }
      if (this.checkBoxManager) {
        this.checkBoxManager.viewport = clonedViewport
      }
      if (this.radioButtonManager) {
        this.radioButtonManager.viewport = clonedViewport
      }
      if (this.pushButtonManager) {
        this.pushButtonManager.viewport = clonedViewport
      }
    }
    this.renderPaintScene()
    this.div = document.createElement("div");
    this.div.className = "annotationContainer";
    this.div.tabIndex = 0;
    this.pageDiv.append(this.div);
    if (this.annotations && this.annotations.length) {
      for (let i = 0; i < this.annotations.length; i++) {
        const annotation = this.annotations[i]
        if (annotation.type === 'line' && annotation.tail !== 'OpenArrow' && !annotation.isDelete) {
          const line = new Line({
            container: this.div,
            annotation,
            page: this.page,
            viewport: this.viewport,
            scale: this.scale,
            eventBus: this.eventBus,
            layer: this
          })
          this.annotationsArray.push(line)
        } else if ((annotation.type === 'square' || annotation.type === 'circle') && !annotation.isDelete) {
          const shape = new Shape({
            container: this.div,
            annotation,
            page: this.page,
            viewport: this.viewport,
            scale: this.scale,
            eventBus: this.eventBus,
            layer: this
          })
          this.annotationsArray.push(shape)
        } else if (annotation.type === 'ink' && !annotation.isDelete) {
          const ink = new Ink({
            container: this.div,
            annotation,
            page: this.page,
            viewport: this.viewport,
            scale: this.scale,
            eventBus: this.eventBus,
            layer: this
          })
          this.annotationsArray.push(ink)
        } else if (annotation.type === 'freetext' && !annotation.isDelete) {
          const freetext = new Freetext({
            container: this.div,
            annotation,
            page: this.page,
            viewport: this.viewport,
            scale: this.scale,
            eventBus: this.eventBus,
            layer: this
          })
          this.annotationsArray.push(freetext)
        }
        if (annotation.type === 'text' && !annotation.isDelete) {
          const text = new Text({
            container: this.div,
            annotation,
            page: this.page,
            viewport: this.viewport,
            scale: this.scale,
            eventBus: this.eventBus,
            layer: this
          })
          this.annotationsArray.push(text)
        }
        if (annotation.type.replace('-', '') === 'textfield' && !annotation.isDelete) {
          const textfield = new TextField({
            container: this.div,
            annotation,
            page: this.page,
            viewport: this.viewport,
            scale: this.scale,
            eventBus: this.eventBus,
            layer: this
          })
          this.annotationsArray.push(textfield)
        }
        if (annotation.type.replace('-', '') === 'checkbox' && !annotation.isDelete) {
          const checkbox = new CheckBox({
            container: this.div,
            annotation,
            page: this.page,
            viewport: this.viewport,
            scale: this.scale,
            eventBus: this.eventBus,
            layer: this
          })
          this.annotationsArray.push(checkbox)
        }
        if (annotation.type.replace('-', '') === 'radiobutton' && !annotation.isDelete) {
          const radiobutton = new RadioButton({
            container: this.div,
            annotation,
            page: this.page,
            viewport: this.viewport,
            scale: this.scale,
            eventBus: this.eventBus,
            layer: this
          })
          this.annotationsArray.push(radiobutton)
        }
        if (annotation.type.replace('-', '') === 'pushbutton' && !annotation.isDelete) {
          const pushbutton = new PushButton({
            container: this.div,
            annotation,
            page: this.page,
            viewport: this.viewport,
            scale: this.scale,
            eventBus: this.eventBus,
            layer: this
          })
          this.annotationsArray.push(pushbutton)
        }
      }
    }

    if (this.tool && this.color) {
      this.handleTool({
        tool: this.tool,
        color: this.color
      })
    }
  }

  cancel() {
    this._cancelled = true;
    this.destroy();
  }

  destroy () {
    this.svgElement.remove()
    this.svgElement = null
    this.div.remove()
    this.div = null
    if (this.freetextManager) {
      this.freetextManager.reset()
    }
    if (this.textManager) {
      this.textManager.reset()
    }
    if (this.textFieldManager) {
      this.textFieldManager.reset()
    }
    if (this.checkBoxManager) {
      this.checkBoxManager.reset()
    }
    if (this.radioButtonManager) {
      this.radioButtonManager.reset()
    }
    if (this.pushButtonManager) {
      this.pushButtonManager.reset()
    }
    if (this.annotateManager) {
      this.annotateManager.reset()
    }
    this.eventBus._off('toolChanged', this.onHandleTool)
    this.eventBus._off('toolModeChanged', this.onHandleToolMode)
    this.eventBus._off('setDefaultSelect', this.onHandleDefaultSelect)
    this.eventBus._off('propertyPanelChanged', this.onHandlePropertyPanelChanged)
  }

  setSceneDimensions (viewport) {

    if (viewport) {
      const clonedViewport = viewport.clone({ dontFlip: true });
      this.viewport = clonedViewport
      if (this.annotateManager) {
        this.annotateManager.viewport = clonedViewport
      }
    }
    if (this.svgElement) {
      const viewport = this.viewport
      const scene = this.scene
      scene.style.width = viewport.width + 'px'
      scene.style.height = viewport.height + 'px'

      this.svgElement.setAttribute("viewBox", `0 0 ${viewport.width} ${viewport.height}`);
    }
  }

  renderPaintScene() {
    if (!this.svgElement) {
      this.svgElement = document.createElementNS("http://www.w3.org/2000/svg", "svg");
      this.scene = this.pageDiv.querySelector('.annotationLayer')
      this.scene.append(this.svgElement)
    }
    this.setSceneDimensions()
  }

  renderAnnotation (annotation, show) {
    if (annotation.type === 'line' || annotation.type === 'arrow') {
      const line = new Line({
        container: this.div,
        annotation,
        page: this.page,
        viewport: this.viewport,
        scale: this.scale,
        eventBus: this.eventBus,
        show: true,
        layer: this
      })
      this.annotationsArray.push(line)
    } else if (annotation.type === 'square' || annotation.type === 'circle') {
      const shape = new Shape({
        container: this.div,
        annotation,
        page: this.page,
        viewport: this.viewport,
        scale: this.scale,
        eventBus: this.eventBus,
        layer: this
      })
      this.annotationsArray.push(shape)
    } else if (annotation.type === 'ink' && !annotation.isDelete) {
      const ink = new Ink({
        container: this.div,
        annotation,
        page: this.page,
        viewport: this.viewport,
        scale: this.scale,
        eventBus: this.eventBus,
        layer: this
      })
      this.annotationsArray.push(ink)
    } else if (annotation.type === 'freetext' && !annotation.isDelete) {
      const freetext = new Freetext({
        container: this.div,
        annotation,
        page: this.page,
        viewport: this.viewport,
        scale: this.scale,
        eventBus: this.eventBus,
        layer: this,
        show
      })
      this.annotationsArray.push(freetext)
    } else if (annotation.type === 'text' && !annotation.isDelete) {
      const text = new Text({
        container: this.div,
        annotation,
        page: this.page,
        viewport: this.viewport,
        scale: this.scale,
        eventBus: this.eventBus,
        layer: this
      })
      this.annotationsArray.push(text)
    } else if (annotation.type === 'text-field' && !annotation.isDelete) {
      const textfield = new TextField({
        container: this.div,
        annotation,
        page: this.page,
        viewport: this.viewport,
        scale: this.scale,
        eventBus: this.eventBus,
        layer: this
      })
      this.annotationsArray.push(textfield)
    } else if (annotation.type === 'check-box' && !annotation.isDelete) {
      const checkbox = new CheckBox({
        container: this.div,
        annotation,
        page: this.page,
        viewport: this.viewport,
        scale: this.scale,
        eventBus: this.eventBus,
        layer: this
      })
      this.annotationsArray.push(checkbox)
    } else if (annotation.type === 'radio-button' && !annotation.isDelete) {
      const radiobutton = new RadioButton({
        container: this.div,
        annotation,
        page: this.page,
        viewport: this.viewport,
        scale: this.scale,
        eventBus: this.eventBus,
        layer: this
      })
      this.annotationsArray.push(radiobutton)
    } else if (annotation.type === 'push-button' && !annotation.isDelete) {
      const pushbutton = new PushButton({
        container: this.div,
        annotation,
        page: this.page,
        viewport: this.viewport,
        scale: this.scale,
        eventBus: this.eventBus,
        layer: this
      })
      this.annotationsArray.push(pushbutton)
    }
  }

  handleToolMode (mode) {
    this.toolMode = mode
    this.pageViewer.toolMode = mode
  }

  handlePropertyPanelChanged (props) {
    if (this.annotations && this.annotations.length) {
      const annotation = this.annotations.find((element) => element.name === this.annotationStore.selectedElementName)
      if (annotation) {
        if (annotation.type.replace('-', '') === 'textfield') {
          this.eventBus.dispatch("textfieldPropertyPanelChanged", props);
        } else if (annotation.type.replace('-', '') === 'checkbox') {
          this.eventBus.dispatch("checkboxPropertyPanelChanged", props);
        } else if (annotation.type.replace('-', '') === 'radiobutton') {
          this.eventBus.dispatch("radioButtonPropertyPanelChanged", props);
        } else if (annotation.type.replace('-', '') === 'pushbutton') {
          this.eventBus.dispatch("pushButtonPropertyPanelChanged", props);
        }
      }
    }
  }
}

export default ComPDFAnnotationLayer;
