import Base from '../annotation/base';
import { MARGIN_DISTANCE } from '../../constants'
import { getActualPoint, getClickPoint, createSvg, createElement } from '../annotation/utils';
import { onClickOutside } from '../ui_utils'

export default class TextField extends Base {
  
  constructor ({
    container,
    annotation,
    page,
    viewport,
    scale,
    eventBus,
    layer,
    show = false
  }) {
    super({
      container,
      annotation,
      page,
      viewport,
      scale,
      eventBus,
      show
    })

    this.layer = layer
    this.hidden = true
    this.initial = false
    this.outline = null

    this.start = null
    this.end = null

    this.newStart = null
    this.newEnd = null

    this.startCircle = null
    this.endCircle = null

    this.show = show
    this.first = true

    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)
    this.onDelete = this.handleDelete.bind(this)
    this.onBlur = this.handleBlur.bind(this)
    this.render()
  }

  render () {
    this.eventBus._on('textfieldPropertyPanelChanged', this.handlePropertyPanel.bind(this))
    this.eventBus._on('setProperty', this.setProperty.bind(this))

    for (let key in this.annotation) {
      if (key === 'background-color') {
        let newName = 'backgroundColor'
        this.annotation[newName] = this.annotation[key]
        delete this.annotation[key]
      }
      if (key === 'fieldname') {
        let newKey = 'fieldName'
        this.annotation[newKey] = this.annotation[key]
        delete this.annotation[key]
      }
      if (key === 'justification') {
        let newKey = 'alignment'
        this.annotation[newKey] = this.annotation[key] === 'center' ? '1' : this.annotation[key] === 'right' ? '2' : '0'
        delete this.annotation[key]
      }
      if (key === 'type' && this.annotation[key] === 'textfield') {
        this.annotation[key] = 'text-field'
      }
      if (key === 'flags') {
        let newKey = 'isHidden'
        this.annotation[newKey] = this.annotation[key] === 'noview' ? '1' : '0'
        delete this.annotation[key]
      }
    }
    if (!this.annotation.value) {
      this.annotation.value = ''
    }
    if (!this.annotation.backgroundColor) {
      this.annotation.backgroundColor = 'transparent'
    }
    if (!this.annotation.alignment) {
      this.annotation.alignment = '0'
    }
    const annotation = this.annotation
    console.log(annotation)

    const { start, end } = this.getActualRect(
      this.viewport,
      this.scale
    )
    this.start = start
    this.end = end

    const rect = this.calculate(start, end)
    const actualbdwidth = (this.annotation.lineWidth || this.annotation.width || 2) * this.scale
    
    const annotationContainer = document.createElement('div')
    annotationContainer.id = annotation.name
    annotationContainer.className = 'annotation'
    annotationContainer.style.top = rect.top + 'px'
    annotationContainer.style.left = rect.left + 'px'
    annotationContainer.style.width = rect.width + 'px'
    annotationContainer.style.height = rect.height + 'px'
    this.annotationContainer = annotationContainer
    this.annotationContainer.addEventListener('mousedown', this.handleClick.bind(this))

    let shapeElement
    shapeElement = createElement(
      'div',
      {
        left: 0,
        top: 0,
        width: `${Math.abs(rect.width)}px`,
        height: `${Math.abs(rect.height)}px`,
        backgroundColor: annotation.backgroundColor || annotation['background-color'],
        position: 'relative'
      }
    )

    this.shapeElement = shapeElement

    this.annotationContainer.append(this.shapeElement)

    this.container.append(this.annotationContainer)

    // create text
    this.setMultiLine(this.annotation.multiline)
    this.shapeElement.append(this.textContainer)
    this.textContainer.addEventListener('blur', this.onBlur)

    this.outerLineContainer = document.createElement('div')
    this.outerLineContainer.className = 'outline-container'
    this.deletetButton = createSvg(
      "svg",
      {
        width: "30",
        height: "30",
        viewBox: "0 0 30 30",
        fill: "none",
        class: 'delete-button'
      },
      {
        width: '30px',
        height: '30px'
      }
    );
    this.deletetButton.addEventListener('click', this.onDelete)
    this.deletetButton.innerHTML = this.deleteSvgStr

    const left = (rect.left + rect.width - 30) + 'px'
    const top = (rect.top + rect.height + 8) + 'px'
    this.deletetButton.style.left = left
    this.deletetButton.style.top = top
    
    const outerLine = document.createElementNS("http://www.w3.org/2000/svg", "svg");

    outerLine.style.position = 'absolute'
    outerLine.style.zIndex = 1
    outerLine.style.left = `${rect.left - 4}px`
    outerLine.style.top = `${rect.top - 4}px`
    outerLine.style.width = `${rect.width + 4 * 2}px`
    outerLine.style.height = `${rect.height + 4 * 2}px`
    this.outerLine = outerLine
    
    this.moveRect = createSvg(
      "rect",
      {
        class: "move",
        'data-id': "move",
        stroke: "#4982E6",
        'stroke-width': 1,
        'fill-opacity': 0,
        width: rect.width,
        height: rect.height,
        x: 4,
        y: 4
      }
    );
    
    this.topLeftRect = createSvg(
      "rect",
      {
        class: "nw-resize",
        'data-id': "nw-resize",
        fill: "#FFFFFF",
        stroke: "#4982E6",
        'stroke-width': 1,
        width: 6,
        height: 6,
        x: 1,
        y: 1
      }
    );

    this.LeftRect = createSvg(
      "rect",
      {
        class: "w-resize",
        'data-id': "w-resize",
        fill: "#FFFFFF",
        stroke: "#4982E6",
        'stroke-width': 1,
        width: 6,
        height: 6,
        x: 1,
        y: rect.height / 2 + 1
      }
    );

    this.bottomLeftRect = createSvg(
      "rect",
      {
        class: "sw-resize",
        'data-id': "sw-resize",
        fill: "#FFFFFF",
        stroke: "#4982E6",
        'stroke-width': 1,
        width: 6,
        height: 6,
        x: 1,
        y: rect.height + 1
      }
    );

    this.bottomRect = createSvg(
      "rect",
      {
        class: "s-resize",
        'data-id': "s-resize",
        fill: "#FFFFFF",
        stroke: "#4982E6",
        'stroke-width': 1,
        width: 6,
        height: 6,
        x: rect.width / 2 + 1,
        y: rect.height + 1
      }
    );
    
    this.bottomRightRect = createSvg(
      "rect",
      {
        class: "se-resize",
        'data-id': "se-resize",
        fill: "#FFFFFF",
        stroke: "#4982E6",
        'stroke-width': 1,
        width: 6,
        height: 6,
        x: rect.width + 1,
        y: rect.height + 1
      }
    );

    this.rightRect = createSvg(
      "rect",
      {
        class: "e-resize",
        'data-id': "e-resize",
        fill: "#FFFFFF",
        stroke: "#4982E6",
        'stroke-width': 1,
        width: 6,
        height: 6,
        x: rect.width + 1,
        y: rect.height / 2 + 1
      }
    );

    this.topRightRect = createSvg(
      "rect",
      {
        class: "ne-resize",
        'data-id': "ne-resize",
        fill: "#FFFFFF",
        stroke: "#4982E6",
        'stroke-width': 1,
        width: 6,
        height: 6,
        x: rect.width + 1,
        y: 1
      }
    );

    this.topRect = createSvg(
      "rect",
      {
        class: "n-resize",
        'data-id': "n-resize",
        fill: "#FFFFFF",
        stroke: "#4982E6",
        'stroke-width': 1,
        width: 6,
        height: 6,
        x: rect.width / 2 + 1,
        y: 1
      }
    );

    this.outerLine.append(this.moveRect)
    this.outerLine.append(this.topLeftRect)
    this.outerLine.append(this.LeftRect)
    this.outerLine.append(this.bottomLeftRect)
    this.outerLine.append(this.bottomRect)
    this.outerLine.append(this.bottomRightRect)
    this.outerLine.append(this.rightRect)
    this.outerLine.append(this.topRightRect)
    this.outerLine.append(this.topRect)
    this.outerLineContainer.append(this.outerLine)
    this.outerLineContainer.append(this.deletetButton)
    this.initial = true

    if (this.layer.toolMode !== 'form') {
      if (this.annotation['isHidden'] === '1') {
        this.annotationContainer.style.display = 'none'
        this.annotationContainer.style.pointerEvents = 'none'
      } else if (this.annotation['isHidden'] === '0') {
        this.annotationContainer.style.display = 'block'
        this.annotationContainer.style.pointerEvents = 'auto'
      }
    }
    if (this.annotation.backgroundColor === 'transparent') {
      this.annotationContainer.style.backgroundColor = '#93B9FD'
    }
    this.eventBus.dispatch('propertyChange', this.annotation)
  }

  getActualRect (viewport, s,) {
    const annotation = this.annotation
    const [x1, y1, x2, y2] = annotation.rect.split(',')
    
    const start = getActualPoint(
      {
        x: x1,
        y: y1
      },
      viewport,
      s
    )
    const end = getActualPoint(
      {
        x: x2,
        y: y2
      },
      viewport,
      s
    )
    return {
      start,
      end
    }
  }

  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    };
  };

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end });
    const actualbdwidth = (this.annotation.width || this.annotation.lineWidth || 2) * this.scale

    return {
      top: initRect.top - actualbdwidth,
      left: initRect.left - actualbdwidth,
      width: initRect.width + actualbdwidth * 2,
      height: initRect.height + actualbdwidth * 2,
    }
  }

  getInitialPoint () {
    const s = this.scale
    const startX = this.start.x
    const startY = this.start.y
    const endX = this.end.x
    const endY = this.end.y

    const { width, height, rotation  } = this.viewport;
    let x1, y1, x2, y2;
    if (rotation === 0) {
      x1 = startX / s;
      y1 = (height - startY) / s; 
      x2 = endX / s;
      y2 = (height - endY) / s;
    } else if (rotation === 90) {
      x1 = startY / s;
      y1 = startX / s;
      x2 = endY / s;
      y2 = endX / s;
    } else if (rotation === 180) {
      x1 =( width - startX) / s;
      y1 = startY / s;
      x2 = (width - endX) / s;
      y2 = endY / s;
    } else {
      x1 = (height - startY) / s
      y1 = (width - startX) / s
      x2 = (height - endY) / s
      y2 = (width - endX)  / s
    }
    return {
      start: {
        x: x1,
        y: y1,
      },
      end: {
        x: x2,
        y: y2,
      },
    };
  }

  updateTool () {
    if (!this.initial) return
    if (this.hidden) {
      if (this.layer.annotationStore.selectedElementName === this.annotation.name) {
        this.layer.annotationStore.selectedElementName = null
      }
      this.outerLineContainer.remove()
    } else {
      if (this.layer.annotationStore.selectedElementName !== this.annotation.name) {
        this.layer.annotationStore.selectedElementName = this.annotation.name
      }
      this.container.append(this.outerLineContainer)
      this.outerLine.addEventListener('mousedown', this.onMousedown)
      this.outerLine.addEventListener('touchstart', this.onMousedown)

      this.eventBus.dispatch('propertyChange', this.annotation)
    }
  }

  handleClick () {
    console.log(this.annotation)
    if (!this.hidden) return
    if (this.layer.toolMode === 'form') {
      this.hidden = false
      this.updateTool()
      
      if (document.getElementsByClassName('right-panel')[0].className.indexOf('closed') !== -1) {
        document.getElementById('propertyPanelButton').click()
      }

      const ignoreList = [this.annotationContainer, this.outerLine, this.deletetButton, document.getElementById('propertyPanelButton'), document.getElementsByClassName('right-panel')[0]]
      onClickOutside(ignoreList, this.handleOutside.bind(this))

      this.textContainer.style.pointerEvents = 'none'
      this.textContainer.style.cursor = 'pointer'
    } else {
      setTimeout(() => {
        this.textContainer.focus();
      }, 0);
      this.textContainer.style.pointerEvents = 'auto'
      this.textContainer.style.cursor = 'text'
    }
  }

  handleOutside () {
    if (!this.hidden && this.layer.annotationStore.selectedElementName === this.annotation.name) {
      document.getElementById('propertyPanelButton').click()
      
      this.eventBus.dispatch('propertyChange', null)
      this.layer.selectedElementName = null
    }

    this.hidden = !this.hidden

    if (this.layer.annotationStore.selectedElementName === this.annotation.name) {
      this.layer.annotationStore.selectedElementName = null
    }
    this.outerLine.removeEventListener('mousedown', this.onMousedown)
    this.outerLine.removeEventListener('touchstart', this.onMousedown)
    this.outerLineContainer.remove()
  }

  handleBlur () {
    if (this.annotation.value !== this.textContainer.value) {
      this.annotation.value = this.textContainer.value
      if (this.annotation.value !== '') {
        this.eventBus.dispatch('annotationChange', {
          type: 'modify',
          annotation: {
            operate: "mod-form",
            name: this.annotation.name,
            page: this.page,
            value: this.annotation.value
          }
        })
      }
    }
    this.textContainer.style.pointerEvents = 'none'
    this.textContainer.style.cursor = 'pointer'
  }

  handleMouseDown (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    const operatorId = event.target.getAttribute('data-id')
  
    const { pageX, pageY } = getClickPoint(event)
    this.startState = {
      operator: operatorId,
      clickX: pageX,
      clickY: pageY,
    }

    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)

    this.handleClick()
  };

  handleMouseMove (event) {
    if (event.type === 'touchmove') {
      document.body.style.overscrollBehavior = 'none';
    }
    this.moving = true
    const { pageX, pageY } = getClickPoint(event)
    const start = this.start
    const end = this.end
    const startState = this.startState

    const { width, height } = this.viewport
    if (startState.operator === 'nw-resize') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.end.y);
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(start.y - 5, y1)
      y1 = Math.max(0, y1)

      this.update({
        start: { x: x1, y: start.y },
        end: { x: end.x, y: y1 },
      });
    } else if (startState.operator === 'w-resize') {
      let x1 = pageX - (startState.clickX - this.start.x);
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)

      this.update({
        start: { x: x1, y: start.y },
        end,
      });
    } else if (startState.operator === 'sw-resize') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.start.y);
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(height, y1)
      y1 = Math.max(end.y + 5, y1)

      this.update({
        start: { x: x1, y: y1 },
        end,
      });
    } else if (startState.operator === 's-resize') {
      let y1 = pageY - (startState.clickY - this.start.y);
      y1 = Math.min(height, y1)
      y1 = Math.max(end.y + 5, y1)
      this.update({
        start: { x: start.x, y: y1 },
        end,
      });
    } else if (startState.operator === 'se-resize') {
      let x = pageX - (startState.clickX - this.end.x);
      let y = pageY - (startState.clickY - this.start.y);
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      y = Math.min(height, y)
      y = Math.max(end.y + 5, y)
      this.update({
        start: { x: start.x, y },
        end: { x, y: end.y },
      });
    } else if (startState.operator === 'e-resize') {
      let x = pageX - (startState.clickX - this.end.x);
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      this.update({
        start,
        end: { x, y: end.y },
      });
    } else if (startState.operator === 'ne-resize') {
      let x = pageX - (startState.clickX - this.end.x);
      let y = pageY - (startState.clickY - this.end.y);
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      y = Math.min(start.y - 5, y)
      y = Math.max(0, y)
      this.update({
        start,
        end: { x, y },
      });
    } else if (startState.operator === 'n-resize') {
      let y = pageY - (startState.clickY - this.end.y);
      y = Math.min(start.y - 5, y)
      y = Math.max(0, y)
      this.update({
        start,
        end: { x: end.x, y },
      });
    } else if (startState.operator === 'move') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.start.y);
      let x2 = pageX - (startState.clickX - this.end.x);
      let y2 = pageY - (startState.clickY - this.end.y);

      const rect = {
        width: Math.abs(start.x -end.x),
        height: Math.abs(start.y -end.y)
      }
      if (x1 < x2) {
        x1 = Math.max(0, x1)
        x1 = Math.min(width - rect.width, x1)
        x2 = Math.max(rect.width, x2)
        x2 = Math.min(width, x2)
      } else {
        x2 = Math.max(0, x2)
        x2 = Math.min(width - rect.width, x2)
        x1 = Math.max(rect.width, x1)
        x1 = Math.min(width, x1)
      }

      if (y1 < y2) {
        y1 = Math.max(0, y1)
        y1 = Math.min(height - rect.height, y1)
        y2 = Math.max(rect.height, y2)
        y2 = Math.min(height, y2)
      } else {
        y2 = Math.max(0, y2)
        y2 = Math.min(height - rect.height, y2)
        y1 = Math.max(rect.height, y1)
        y1 = Math.min(height, y1)
      }

      this.update({
        start: { x: x1, y: y1 },
        end: { x: x2, y: y2 },
      });
    }
  }

  handleDelete (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    this.handleOutside()
    this.annotationContainer.remove()

    this.annotation.isDelete = true
    this.eventBus.dispatch('annotationChange', {
      type: 'delete',
      annotation: {
        operate: "del-annot",
        name: this.annotation.name,
        page: this.page
      }
    })
  }

  handleMouseUp (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    if (event.type === 'touchend') {
      document.body.style.overscrollBehavior = 'auto';
    }
    this.moving = false
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
    
    const { pageX, pageY } = getClickPoint(event)
    if (pageX === this.startState.clickX && pageY === this.startState.clickY) return

    this.start = this.newStart
    this.end = this.newEnd

    const annotation = this.annotation
    const { start, end } = this.getInitialPoint()

    const linePoints = start.x + ',' + start.y + ',' + end.x + ',' + end.y
    if (linePoints === annotation.linePoints) return
    annotation.linePoints = linePoints

    this.eventBus.dispatch('annotationChange', {
      type: 'modify',
      annotation: {
        operate: "mod-form",
        name: annotation.name,
        page: this.page,
        rect: linePoints
      }
    })
  }

  update ({ start, end }) {
    const rect = this.calculate(start, end)

    this.newStart = start
    this.newEnd = end

    const actualbdwidth = (this.annotation.lineWidth || this.annotation.width || 2) * this.scale

    this.annotationContainer.style.top = rect.top + 'px'
    this.annotationContainer.style.left = rect.left + 'px'
    this.annotationContainer.style.width = rect.width + 'px'
    this.annotationContainer.style.height = rect.height + 'px'
    
    this.annotationContainer.style.width = `${rect.width}px`;
    this.annotationContainer.style.height = `${rect.height}px`;

    this.shapeElement.style.width = `${rect.width}px`;
    this.shapeElement.style.height = `${rect.height}px`;

    this.outerLine.style.left = `${rect.left - 4}px`
    this.outerLine.style.top = `${rect.top - 4}px`
    this.outerLine.style.width = `${rect.width + 4 * 2}px`
    this.outerLine.style.height = `${rect.height + 4 * 2}px`

    const left = (rect.left + rect.width - 30) + 'px'
    const top = (rect.top + rect.height + 8) + 'px'
    this.deletetButton.style.left = left
    this.deletetButton.style.top = top
    
    this.moveRect.setAttribute('width',  rect.width)
    this.moveRect.setAttribute('height',  rect.height)
    
    this.LeftRect.setAttribute("y", rect.height / 2 + 1)

    this.bottomLeftRect.setAttribute("y", rect.height + 1)

    this.bottomRect.setAttribute("x", rect.width / 2 + 1)
    this.bottomRect.setAttribute("y", rect.height + 1)


    this.bottomRightRect.setAttribute("x", rect.width + 1)
    this.bottomRightRect.setAttribute("y", rect.height + 1)


    this.rightRect.setAttribute("x", rect.width + 1)
    this.rightRect.setAttribute("y", rect.height / 2 + 1)

    
    this.topRightRect.setAttribute("x", rect.width + 1)

    this.topRect.setAttribute("x", rect.width / 2 + 1)
  }

  handlePropertyPanel (props) {
    if (this.layer.annotationStore.selectedElementName !== this.annotation.name) return

    let updatePropsNum = 0
    for (const key in props) {
      for (const item in this.annotation) {
        if (item === key && this.annotation[item] !== props[key]) {
          if (props['value'] !== '') {
            this.annotation[item] = props[key]
            updatePropsNum ++
          }

          if (key === 'backgroundColor') {
            this.shapeElement.style[key] = props[key]
          }
          if (key === 'alignment') {
            let alignment = 'left'
            if (props[key] === '0') {
              alignment = 'left'
            } else if (props[key] === '1') {
              alignment = 'center'
            } else if (props[key] === '2') {
              alignment = 'right'
            }
            this.textContainer.style['textAlign'] = alignment
          }
          if (key === 'fontSize') {
            this.textContainer.style.fontSize = props[key] + 'px'
          }
          if (key === 'textColor') {
            this.textContainer.style.color = props[key]
          }
          if (key === 'fontName') {
            this.textContainer.style.fontFamily = props[key]
          }
          if (key === 'value') {
            const nodeName = this.textContainer.nodeName
            if (nodeName === 'INPUT') {
              this.textContainer.setAttribute(key, props[key])
            } else {
              this.textContainer.innerHTML = this.annotation.value
            }
          }
          if (key === 'multiline') {
            this.changeElementMultiLine(props[key])
          }
        }
      }
    }

    if (updatePropsNum > 0) {
      const annotation = {
        operate: "mod-form",
        name: this.annotation.name,
        page: this.page
      }
      const updateAnnot = Object.assign({}, annotation, props)
      this.eventBus.dispatch('annotationChange', {
        type: 'modify',
        annotation: updateAnnot
      })
    }
  }

  changeElementMultiLine (boolean) {
    this.textContainer.remove()
    this.textContainer = null
    this.setMultiLine(boolean)
    this.shapeElement.append(this.textContainer)
    this.textContainer.removeEventListener('blur', this.onBlur)
    this.textContainer.addEventListener('blur', this.onBlur)
  }

  setMultiLine (boolean) {
    let textContainer = null
    if (boolean) {
      textContainer = createElement(
        'textarea',
        {
          width: '100%',
          height: '100%',
          outline: 'none',
          border: 'none',
          color: this.annotation.textColor,
          backgroundColor: 'transparent',
          pointerEvents: 'none',
          cursor: 'pointer',
          resize: 'none',
          textAlign: this.annotation.alignment === '1' ? 'center' :this.annotation.alignment === '2' ? 'right' : 'left',
          fontSize: this.annotation.fontSize,
          fontFamily: this.annotation.fontName
        }
      )
      textContainer.innerHTML = this.annotation.value
    } else {
      textContainer = createElement(
        'input',
        {
          width: '100%',
          height: '100%',
          outline: 'none',
          border: 'none',
          color: this.annotation.textColor,
          backgroundColor: 'transparent',
          pointerEvents: 'none',
          cursor: 'pointer',
          textAlign: this.annotation.alignment === '1' ? 'center' :this.annotation.alignment === '2' ? 'right' : 'left',
          fontSize: this.annotation.fontSize,
          fontFamily: this.annotation.fontName
        },
        { value: this.annotation.value }
      )
    }
    this.textContainer = textContainer
  }

  setProperty (props) {
    if (props.name !== this.annotation.name) return
    console.log(props, this.annotation)

    for (const key in props) {
      for (const item in this.annotation) {
        if (item === key && this.annotation[item] !== props[key]) {
          this.annotation[item] = props[key]
          
          if (key === 'backgroundColor') {
            this.shapeElement.style[key] = props[key]
          }
          if (key === 'alignment') {
            let alignment = 'left'
            if (props[key] === '0') {
              alignment = 'left'
            } else if (props[key] === '1') {
              alignment = 'center'
            } else if (props[key] === '2') {
              alignment = 'right'
            }
            this.textContainer.style['textAlign'] = alignment
          }
          if (key === 'fontSize') {
            this.textContainer.style.fontSize = props[key] + 'px'
          }
          if (key === 'textColor') {
            this.textContainer.style.color = props[key]
          }
          if (key === 'fontName') {
            this.textContainer.style.fontFamily = props[key]
          }
          if (key === 'value') {
            const nodeName = this.textContainer.nodeName
            if (nodeName === 'INPUT') {
              this.textContainer.setAttribute(key, props[key])
            } else {
              this.textContainer.innerHTML = this.annotation.value
            }
          }
          if (key === 'multiline') {
            this.changeElementMultiLine(props[key])
          }
        }
      }
    }

    const annotation = {
      operate: "mod-form",
      page: this.page
    }
    const updateAnnot = Object.assign({}, annotation, props)
    this.eventBus.dispatch('annotationChange', {
      type: 'modify',
      annotation: updateAnnot
    })
  }
}
